import contextlib
import enum
import functools
import random
import typing as t
import uuid
from collections import abc
from dataclasses import dataclass
from functools import cache

import diwrappers._commons._data as d


@dataclass
class Injector(t.Generic[d.Data]):
    """
    A dependency injection container.

    This class provides a flexible way to manage dependencies in your
    application, supporting both regular dependency injection and
    testing scenarios through context managers that allow temporary
    dependency replacement.

    """

    _constructor: t.Callable[[], d.Data]
    """Function that creates new instances of the dependency."""

    @contextlib.contextmanager
    def fake_value(self, val: d.Data) -> abc.Generator[d.Data, None, None]:
        """
        Temporarily replace the dependency with a specific value.

        Args:
            val: The value to use instead of the normal dependency.

        Yields:
            The provided fake value.

        """
        tmp_constructor = self._constructor
        self._constructor = lambda: val
        try:
            yield val
        finally:
            self._constructor = tmp_constructor

    def faker(self, fake_constructor: t.Callable[[], d.Data]):
        """
        Create a context manager to replace the dependency constructor.

        Args:
            fake_constructor:
                A callable that will temporarily replace the normal
                dependency constructor.

        Returns:
            A context manager that can be used to temporarily
            replace the dependency constructor.

        """

        @contextlib.contextmanager
        def wrapper():
            tmp_constructor = self._constructor
            self._constructor = fake_constructor
            try:
                yield
            finally:
                self._constructor = tmp_constructor

        return wrapper

    def inject(
        self,
        task: t.Callable[t.Concatenate[d.Data, d.TaskParams], d.TaskReturn],
    ) -> t.Callable[d.TaskParams, d.TaskReturn]:
        """
        Decorate a function to inject the dependency as its first argument.

        Args:
            task:
                The function to be decorated.
                Its first parameter must be of type d.Data.

        Returns:
            A wrapped function that will automatically
            receive the dependency as its first argument.

        """

        @functools.wraps(task)
        def _wrapper(*args: d.TaskParams.args, **kwargs: d.TaskParams.kwargs):
            """Create and inject the dependency."""
            data = self._constructor()
            return task(data, *args, **kwargs)

        return _wrapper


def dependency(func: t.Callable[[], d.Data]) -> Injector[d.Data]:
    """
    Create a dependency injector from a constructor function.

    Args:
        func: A constructor function that creates the dependency

    """
    return Injector(func)


if d.is_test_env():  # pragma: no cover
    import pytest
    # SECTION tests

    # fake data
    NAME = "user_name"
    PROD_TOKEN = uuid.uuid4().hex
    FAKE_TOKEN = uuid.uuid4().hex
    PROD_URL = "http://prod-api.com"
    FAKE_URL = "http://fake-api.com"
    FAKE_INT = 1234

    def test_token_injection() -> None:
        @dependency
        def token() -> str:
            return "test_token"

        @token.inject
        def build_http_headers(token: str):
            return {"Authorization": f"Bearer {token}"}

        for i in range(3):
            headers = build_http_headers()
            assert headers["Authorization"] == "Bearer test_token", (
                f"Attempt {i}"
            )

    def test_singleton_dependency() -> None:
        counter = 0

        @dependency
        @cache
        def get_counter():
            nonlocal counter
            counter += 1
            return counter

        @get_counter.inject
        def read_counter(counter: int):
            return counter

        assert read_counter() == 1, "must always return the same value"
        assert read_counter() == 1, "must always return the same value"
        assert read_counter() == 1, "must always return the same value"
        assert read_counter() == 1, "must always return the same value"

        assert counter == 1, (
            "constructor can only be called once"
        )  # Constructor called only once

    # types and data for using random during tests

    class _NormalRange(enum.IntEnum):
        """Ground truth range for random number generation."""

        START = 1
        END = 10

    class _TestRAnge(enum.IntEnum):
        """Modified range for testing purposes."""

        START = 11
        END = 15

    N_TRIALS = 100
    """ Number of times the distribution will be sampled """

    SEED = 42
    """ Seed for the pRNG """

    @pytest.fixture(autouse=True)
    def set_random_seed() -> None:
        random.seed(SEED)

    def test_faker_decorator() -> None:
        @dependency
        def random_int() -> int:
            return random.randint(_NormalRange.START, _NormalRange.END)  # nosec - for testing purposes, not used in package

        @random_int.faker
        def fake_random_int() -> int:
            return random.randint(_TestRAnge.START, _TestRAnge.END)  # nosec - for testing purposes, not used in package

        @random_int.inject
        def get_number(random_int: int) -> int:
            return random_int

        # Test normal behavior
        assert all(
            _NormalRange.START <= get_number() <= _NormalRange.END
            for _ in range(N_TRIALS)
        )

        # Test with faker
        with fake_random_int():
            assert all(
                _TestRAnge.START <= get_number() <= _TestRAnge.END
                for _ in range(N_TRIALS)
            )

        # Test restoration after context
        assert all(
            _NormalRange.START <= get_number() <= _NormalRange.END
            for _ in range(N_TRIALS)
        )

    def test_fake_value_context() -> None:
        @dependency
        def random_int():
            return random.randint(_NormalRange.START, _NormalRange.END)  # nosec - for testing purposes, not used in package

        @random_int.inject
        def get_number(random_int: int):
            return random_int

        # Test normal behavior
        assert all(
            _NormalRange.START <= get_number() <= _NormalRange.END
            for _ in range(N_TRIALS)
        )

        # Test with fake value
        with random_int.fake_value(FAKE_INT) as fake_int:
            assert get_number() == FAKE_INT
            assert fake_int == FAKE_INT

        # Test restoration after context
        assert all(
            _NormalRange.START <= get_number() <= _NormalRange.END
            for _ in range(N_TRIALS)
        )

    def test_multiple_fake_contexts() -> None:
        @dependency
        def random_int():
            return random.randint(_NormalRange.START, _NormalRange.END)

        @dependency
        def token():
            return PROD_TOKEN

        @dependency
        def api_base_url():
            return PROD_URL

        @random_int.inject
        @token.inject
        @api_base_url.inject
        def get_random_user(
            base_url: str,
            token: str,
            random_int: int,
            name: str,
        ):
            return base_url, token, random_int, name

        with (
            random_int.fake_value(FAKE_INT) as fake_int,
            token.fake_value(FAKE_TOKEN) as fake_token,
            api_base_url.fake_value(FAKE_URL) as fake_api_base_url,
        ):
            assert fake_int == FAKE_INT
            assert fake_token == FAKE_TOKEN
            assert fake_api_base_url == FAKE_URL

            _base_url, _token, _random_int, _name = get_random_user(name=NAME)

            assert _base_url == fake_api_base_url
            assert _token == fake_token
            assert _random_int == fake_int
            assert _name == NAME

    def test_chained_dependencies() -> None:
        @dependency
        def token() -> str:
            return "test_token"

        values: list[str] = []

        @dependency
        @token.inject
        def client(token: str) -> str:
            values.append(token)
            return "test_client"

        @client.inject
        def use_client(client: str):
            values.append(client)
            return client

        result = use_client()
        assert result == "test_client"
        assert values == ["test_token", "test_client"]

    def test_multiple_dependencies() -> None:
        @dependency
        def logger() -> str:
            return "logger_instance"

        @dependency
        def db_connection() -> str:
            return "db_connection_instance"

        @logger.inject
        @db_connection.inject
        def use_services(db_connection: str, logger: str) -> str:
            return f"Using {db_connection} with {logger}"

        assert (
            use_services()
            == "Using db_connection_instance with logger_instance"
        )

    def test_dependency_replacement() -> None:
        @dependency
        def config() -> dict[str, str]:
            return {"env": "production"}

        @config.inject
        def get_env(config: dict[str, str]) -> str:
            return config["env"]

        assert get_env() == "production"

        with config.fake_value({"env": "test"}):
            assert get_env() == "test"

        assert get_env() == "production"

    def test_injected_function_exception() -> None:
        @dependency
        def db_connection() -> str:
            return "db"

        @db_connection.inject
        def failing_function(_db_connection: str) -> t.NoReturn:
            msg = "Simulated error"
            raise ValueError(msg)

        with pytest.raises(ValueError, match="Simulated error"):
            failing_function()

    def test_thread_safety() -> None:
        import threading

        range_start = 1
        range_end = 100

        @dependency
        def random_number() -> int:
            return random.randint(range_start, range_end)

        @random_number.inject
        def get_number(random_number: int) -> int:
            return random_number

        results: list[int] = []

        def worker() -> None:
            results.append(get_number())

        number_of_threads = 10
        threads = [
            threading.Thread(target=worker) for _ in range(number_of_threads)
        ]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        assert len(results) == number_of_threads
        assert all(
            isinstance(num, int) and range_start <= num <= range_end
            for num in results
        )

    GT_TOKEN = uuid.uuid4().hex
    FAKE_1 = uuid.uuid4().hex
    FAKE_2 = uuid.uuid4().hex

    def test_nested_fakers() -> None:
        @dependency
        def token() -> str:
            return GT_TOKEN

        @token.faker
        def fake_token_1():
            return FAKE_1

        @token.faker
        def fake_token_2():
            return FAKE_2

        @token.inject
        def assert_gt(token: str) -> None:
            assert token == GT_TOKEN

        @token.inject
        def assert_fake_1(token: str) -> None:
            assert token == FAKE_1

        @token.inject
        def assert_fake_2(token: str) -> None:
            assert token == FAKE_2

        assert_gt()

        with fake_token_1():
            assert_fake_1()
            with fake_token_2():
                assert_fake_2()
                with fake_token_1():
                    assert_fake_1()
                assert_fake_2()

                assert_fake_2()
                with fake_token_1():
                    assert_fake_1()
                assert_fake_2()

            assert_fake_1()

        assert_gt()



class DependencyInjectionError(Exception):
    """Base exception for all dependency injection related errors."""


class DependencyLeakError(DependencyInjectionError):
    """Raised when a dependency is returned or leaked from its context."""

    def __init__(self) -> None:
        super().__init__(
            "Dependency cannot be returned or leaked from its context",
        )


class MissingContextError(DependencyInjectionError):
    """Raised when trying to inject a dependency without an ensure context."""

    def __init__(self) -> None:
        super().__init__(
            "Dependency injection requires an ensure context"
            "- please use ensure decorator",
        )



"""Entrypoint of diwrappers."""

from ._async_contextual_dependency import async_contextual_dependency
from ._async_dependency import async_dependency
from ._commons._exceptions import (
    DependencyInjectionError,
    DependencyLeakError,
    MissingContextError,
)
from ._contextual_dependency import contextual_dependency
from ._dependency import dependency

__all__ = [
    "DependencyInjectionError",
    "DependencyLeakError",
    "MissingContextError",
    "async_contextual_dependency",
    "async_dependency",
    "contextual_dependency",
    "dependency",
]



import os
import typing as t

Data = t.TypeVar("Data")
""" The type of the dependency being managed by this injector."""

TaskReturn = t.TypeVar("TaskReturn")
""" Type parameters for the decorated function's arguments"""

TaskParams = t.ParamSpec("TaskParams")
"""Return type of the decorated function"""


def is_tuple(val: object) -> t.TypeGuard[tuple[object]]:
    """
    Check if a value is a tuple.

    Examples:
        >>> is_tuple((1, 2, 3))
        True

        >>> is_tuple([1, 2, 3])
        False

        >>> is_tuple(())
        True

        >>> is_tuple("not a tuple")
        False

        >>> is_tuple(tuple(range(3)))
        True

        >>> is_tuple(tuple())
        True

        >>> is_tuple(dict(a=1).items())  # dict_items is not a tuple
        False

    """
    return isinstance(val, tuple)


def is_list(val: object) -> t.TypeGuard[list[object]]:
    """
    Check if a value is a list.

    Examples:
        >>> is_list([1, 2, 3])
        True

        >>> is_list((1, 2, 3))
        False

        >>> is_list([])
        True

        >>> is_list("not a list")
        False

        >>> is_list(list(range(3)))
        True

        >>> is_list(list())
        True

        >>> is_list(dict(a=1).keys())  # dict_keys is not a list
        False

    """
    return isinstance(val, list)


def is_dict(val: object) -> t.TypeGuard[dict[object, object]]:
    """
    Check if a value is a dictionary.

    Examples:
        >>> is_dict({"a": 1, "b": 2})
        True

        >>> is_dict([("a", 1), ("b", 2)])
        False

        >>> is_dict({})
        True

        >>> is_dict("not a dict")
        False

        >>> is_dict(dict(a=1, b=2))
        True

        >>> is_dict(dict())
        True

        >>> class DictLike:
        ...     def __getitem__(self, key): return None
        ...     def __setitem__(self, key, value): pass
        >>> is_dict(DictLike())  # dict-like object but not a dict
        False

    """
    return isinstance(val, dict)


MAX_DEPTH = 5


def contains_value(needle: object, haystack: object, depth: int = 1) -> bool:
    """
    Check if needle exists within haystack, including in nested structures.

    Examples:
        >>> contains_value(5, 5)
        True

        >>> contains_value("test", "different")
        False

        >>> contains_value(42, [1, 2, [3, 4, [42]]])
        True

        >>> contains_value("x", {"a": 1, "b": {"c": "x"}})
        True

        >>> contains_value("missing", [1, 2, 3])
        False

        >>> contains_value(True, {"a": False, "b": [{"c": True}]})
        True

        >>> class TestClass:
        ...     def __init__(self):
        ...         self.value = ["hidden"]
        >>> obj = TestClass()
        >>> contains_value("hidden", obj)
        True

        >>> contains_value(None, [1, None, 3])
        True

        >>> contains_value("key", {"key": "value"})
        True

    """
    if needle == haystack:
        return True

    if isinstance(haystack, (int, str, bool)) or depth == MAX_DEPTH:
        return False

    depth = depth + 1

    if is_tuple(haystack) or is_list(haystack):
        return any(
            contains_value(needle, item, depth=depth) for item in haystack
        )

    if is_dict(haystack):
        return any(
            contains_value(needle, k, depth=depth)
            or contains_value(needle, v, depth=depth)
            for k, v in haystack.items()
        )

    if hasattr(haystack, "__dict__"):
        return contains_value(needle, vars(haystack), depth=depth)

    return False


TEST_VAR_NAME = "DIWRAPPERS_TEST"
""" Env variable that indicates if this is a test run """


def is_test_env():
    return TEST_VAR_NAME in os.environ and os.environ[TEST_VAR_NAME] == "true"



from __future__ import annotations

import contextlib
import functools
import typing as t
from dataclasses import dataclass

import diwrappers._commons._data as d
import diwrappers._commons._exceptions as e

ContextualConstructor = t.Callable[
    [],
    contextlib.AbstractContextManager[d.Data],
]

P = t.ParamSpec("P")
R = t.TypeVar("R")


@dataclass
class ContextualInjector(t.Generic[d.Data]):
    """A dependency injector that manages contextual dependencies."""

    _constructor: ContextualConstructor[d.Data]
    """Function that creates new instances of the dependency."""

    _data: d.Data | None = None

    def ensure(self, fn: t.Callable[P, R]):
        """Ensure that the dependency is available within the function scope."""

        def wrapper(*args: P.args, **kwargs: P.kwargs):
            with self._constructor() as data:
                self._data = data
                try:
                    res = fn(*args, **kwargs)
                except:
                    self._data = None
                    raise
                if d.contains_value(needle=data, haystack=res):
                    raise e.DependencyLeakError
                self._data = None
            return res

        return wrapper

    def inject(
        self,
        task: t.Callable[t.Concatenate[d.Data, d.TaskParams], d.TaskReturn],
    ) -> t.Callable[d.TaskParams, d.TaskReturn]:
        @functools.wraps(task)
        def _wrapper(*args: d.TaskParams.args, **kwargs: d.TaskParams.kwargs):
            """Create and inject the dependency."""
            if self._data is None:
                raise e.MissingContextError

            return task(self._data, *args, **kwargs)

        return _wrapper


def contextual_dependency(
    func: ContextualConstructor[d.Data],
) -> ContextualInjector[d.Data]:
    return ContextualInjector(func)


if d.is_test_env():  # pragma: no cover
    import pytest

    def test_missing_context_error():
        @contextual_dependency
        @contextlib.contextmanager
        def get_db():
            yield "db_connection"

        @get_db.inject
        def use_db(db: str):
            return "using " + db

        with pytest.raises(e.MissingContextError):
            use_db()

    def test_dependency_leak_error():
        @contextual_dependency
        @contextlib.contextmanager
        def get_secret():
            yield "secret_value"

        @get_secret.ensure
        @get_secret.inject
        def leak_secret(secret: str):
            return secret

        with pytest.raises(e.DependencyLeakError):
            leak_secret()

        @get_secret.ensure
        @get_secret.inject
        def wrap_secret(secret: str):
            return f"prefix_{secret}"

        assert wrap_secret() == "prefix_secret_value"

    def test_nested_ensure_contexts():
        @contextual_dependency
        @contextlib.contextmanager
        def get_secret():
            yield "secret"

        @get_secret.ensure
        def outer():
            assert inner() == "using secret"

            @get_secret.ensure
            def nested():
                return inner()

            return nested()

        @get_secret.inject
        def inner(secret: str):
            return f"using {secret}"

        assert outer() == "using secret"

    def test_multiple_injections_same_context():
        @contextual_dependency
        @contextlib.contextmanager
        def get_db():
            yield "db"

        @get_db.ensure
        def process_data():
            res1 = func1()
            res2 = func2()
            return res1 + res2

        @get_db.inject
        def func1(db: str):
            return f"1_{db}"

        @get_db.inject
        def func2(db: str):
            return f"2_{db}"

        assert process_data() == "1_db2_db"

    def test_context_cleanup():
        cleanup_happened = False

        @contextual_dependency
        @contextlib.contextmanager
        def get_resource():
            nonlocal cleanup_happened
            yield "resource"
            cleanup_happened = True

        @get_resource.ensure
        def use_resource():
            return "ok"

        use_resource()
        assert cleanup_happened, "Context should clean up after use"

    def test_exception_during_execution():
        entered_context = False
        exited_context = False

        @contextual_dependency
        @contextlib.contextmanager
        def resource():
            nonlocal entered_context
            entered_context = True

            try:
                yield "resource"
            finally:
                nonlocal exited_context
                exited_context = True

        @resource.ensure
        @resource.inject
        def failing_function(_resource: str):
            msg = "Something went wrong"
            raise ValueError(msg)

        with pytest.raises(ValueError, match="Something went wrong"):
            failing_function()

        assert entered_context, "Context should be entered"
        assert exited_context, (
            "Context should be cleaned up even after exception"
        )

    def test_complex_dependency_leak():
        @contextual_dependency
        @contextlib.contextmanager
        def get_secret():
            yield "secret"

        @get_secret.ensure
        @get_secret.inject
        def leak_in_dict(secret: str):
            return {"data": secret}  # Leak in dict

        @get_secret.ensure
        @get_secret.inject
        def leak_in_list(secret: str):
            return [1, 2, secret]  # Leak in list

        @get_secret.ensure
        @get_secret.inject
        def leak_in_nested(secret: str):
            return {"outer": {"inner": [1, {"secret": secret}]}}  # Nested leak

        for func in [leak_in_dict, leak_in_list, leak_in_nested]:
            with pytest.raises(e.DependencyLeakError):
                func()

    def test_ensure_decorator_return_types():
        @contextual_dependency
        @contextlib.contextmanager
        def get_db():
            yield "db"

        # Test with different return types
        @get_db.ensure
        def return_none():
            return None

        number = 42

        @get_db.ensure
        def return_int():
            return number

        @get_db.ensure
        def return_dict():
            return {"safe": "value"}

        assert return_none() is None
        assert return_int() == number
        assert return_dict() == {"safe": "value"}

    def test_ensure_without_injection():
        @contextual_dependency
        @contextlib.contextmanager
        def get_db():
            yield "db"

        # ensure can be used without inject
        @get_db.ensure
        def standalone():
            return "ok"

        assert standalone() == "ok"


if __name__ == "__main__":  # pragma: no cover

    @contextual_dependency
    @contextlib.contextmanager
    def db_conn():
        yield 1234

    @db_conn.inject
    def do_work(db_conn: int):
        return db_conn

    @db_conn.ensure
    def some_other_function(): ...

    def main():
        return do_work()

    res = main()



from __future__ import annotations

import contextlib
import functools
import typing as t
from dataclasses import dataclass

import diwrappers._commons._data as d
import diwrappers._commons._exceptions as e

AsyncContextualConstructor = t.Callable[
    [],
    contextlib.AbstractAsyncContextManager[d.Data],
]

P = t.ParamSpec("P")
R = t.TypeVar("R")


@dataclass
class AsyncContextualInjector(t.Generic[d.Data]):
    """A dependency injector that manages contextual dependencies."""

    _constructor: AsyncContextualConstructor[d.Data]
    """Function that creates new instances of the dependency."""

    _data: d.Data | None = None

    def ensure(self, fn: t.Callable[P, t.Awaitable[R]]):
        """Ensure that the dependency is available within the function scope."""

        async def wrapper(*args: P.args, **kwargs: P.kwargs):
            async with self._constructor() as data:
                self._data = data
                res = await fn(*args, **kwargs)
                if d.contains_value(needle=data, haystack=res):
                    raise e.DependencyLeakError
                self._data = None
            return res

        return wrapper

    def inject(
        self,
        task: t.Callable[
            t.Concatenate[d.Data, d.TaskParams], t.Awaitable[d.TaskReturn]
        ],
    ) -> t.Callable[d.TaskParams, t.Awaitable[d.TaskReturn]]:
        @functools.wraps(task)
        async def _wrapper(
            *args: d.TaskParams.args,
            **kwargs: d.TaskParams.kwargs,
        ):
            """Create and inject the dependency."""
            if self._data is None:
                raise e.MissingContextError

            return await task(self._data, *args, **kwargs)

        return _wrapper


def async_contextual_dependency(
    func: AsyncContextualConstructor[d.Data],
) -> AsyncContextualInjector[d.Data]:
    return AsyncContextualInjector(func)


if __name__ == "__main__":  # pragma: no cover

    @async_contextual_dependency
    @contextlib.asynccontextmanager
    async def db_conn():
        yield 1234

    @db_conn.inject
    async def do_work(db_conn: int):
        return db_conn

    @db_conn.ensure
    async def some_other_function(): ...

    async def main():
        return await do_work()

    res = main()

if d.is_test_env():  # pragma: no cover
    import pytest

    @pytest.mark.asyncio
    async def test_missing_context_error():
        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_db():
            yield "db_connection"

        @get_db.inject
        async def use_db(db: str):
            return "using " + db

        # Try to use without ensure context
        with pytest.raises(e.MissingContextError):
            await use_db()

    @pytest.mark.asyncio
    async def test_dependency_leak_error():
        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_secret():
            yield "secret_value"

        @get_secret.ensure
        @get_secret.inject
        async def leak_secret(secret: str):
            return secret

        with pytest.raises(e.DependencyLeakError):
            await leak_secret()

        @get_secret.ensure
        @get_secret.inject
        async def wrap_secret(secret: str):
            return f"prefix_{secret}"

        assert await wrap_secret() == "prefix_secret_value"

    @pytest.mark.asyncio
    async def test_nested_ensure_contexts():
        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_resource():
            yield "resource"

        @get_resource.ensure
        async def outer():
            # Should work - context exists
            assert await inner() == "using resource"

            @get_resource.ensure
            async def nested():
                return await inner()

            return await nested()

        @get_resource.inject
        async def inner(resource: str):
            return f"using {resource}"

        assert await outer() == "using resource"

    @pytest.mark.asyncio
    async def test_cleanup():
        cleanup_called = False

        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_resource():
            try:
                yield "resource"
            finally:
                nonlocal cleanup_called
                cleanup_called = True

        @get_resource.ensure
        async def use_resource():
            return "ok"

        await use_resource()
        assert cleanup_called, "Cleanup should be called"

    @pytest.mark.asyncio
    async def test_exception_handling():
        cleanup_called = False

        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_resource():
            try:
                yield "resource"
            finally:
                nonlocal cleanup_called
                cleanup_called = True

        @get_resource.ensure
        @get_resource.inject
        async def failing_function(_resource: str):
            msg = "Test error"
            raise ValueError(msg)

        with pytest.raises(ValueError, match="Test error"):
            await failing_function()

        assert cleanup_called, "Cleanup should be called even after exception"

    @pytest.mark.asyncio
    async def test_complex_dependency_leak():
        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_secret():
            yield "secret"

        @get_secret.ensure
        @get_secret.inject
        async def leak_in_dict(secret: str):
            return {"data": secret}

        @get_secret.ensure
        @get_secret.inject
        async def leak_in_list(secret: str):
            return [1, secret, 3]

        @get_secret.ensure
        @get_secret.inject
        async def leak_in_nested(secret: str):
            return {"outer": {"inner": [1, {"secret": secret}]}}

        for func in [leak_in_dict, leak_in_list, leak_in_nested]:
            with pytest.raises(e.DependencyLeakError):
                await func()

    @pytest.mark.asyncio
    async def test_multiple_injections():
        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_db():
            yield "db"

        @get_db.ensure
        async def process_data():
            res1 = await func1()
            res2 = await func2()
            return [res1, res2]

        @get_db.inject
        async def func1(db: str):
            return f"1_{db}"

        @get_db.inject
        async def func2(db: str):
            return f"2_{db}"

        result = await process_data()
        assert result == ["1_db", "2_db"]

    @pytest.mark.asyncio
    async def test_async_context_timing():
        events: list[str] = []

        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_resource():
            events.append("enter")
            yield "resource"
            events.append("exit")

        @get_resource.ensure
        @get_resource.inject
        async def use_resource(resource: str):
            events.append("use")
            return resource + "abc"

        await use_resource()
        assert events == ["enter", "use", "exit"]

    @pytest.mark.asyncio
    async def test_injection_with_parameters():
        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_prefix():
            yield "test"

        @get_prefix.ensure
        @get_prefix.inject
        async def format_string(prefix: str, name: str, count: int = 1):
            return f"{prefix}-{name}-{count}"

        result = await format_string("user", count=5)
        assert result == "test-user-5"

    @pytest.mark.asyncio
    async def test_ensure_without_inject():
        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def get_db():
            yield "db"

        # ensure without inject should work
        @get_db.ensure
        async def standalone():
            return "ok"

        assert await standalone() == "ok"

    @pytest.mark.asyncio
    async def test_multiple_resources():
        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def db():
            yield "db"

        @async_contextual_dependency
        @contextlib.asynccontextmanager
        async def cache():
            yield "cache"

        @db.ensure
        @cache.ensure
        @db.inject
        @cache.inject
        async def use_both(cache: str, db: str):
            return f"{cache}-{db}"

        assert await use_both() == "cache-db"



import contextlib
import functools
import typing as t
from collections import abc
from dataclasses import dataclass

import diwrappers._commons._data as d


@dataclass
class AsyncInjector(t.Generic[d.Data]):
    _constructor: t.Callable[[], abc.Awaitable[d.Data]]
    """Function that creates new instances of the dependency."""

    @contextlib.contextmanager
    def fake_value(self, val: d.Data):
        tmp_constructor = self._constructor

        async def new_constructor():
            return val

        self._constructor = new_constructor
        try:
            yield val
        finally:
            self._constructor = tmp_constructor

    def faker(self, fake_constructor: t.Callable[[], abc.Awaitable[d.Data]]):
        @contextlib.contextmanager
        def wrapper():
            tmp_constructor = self._constructor
            self._constructor = fake_constructor
            try:
                yield
            finally:
                self._constructor = tmp_constructor

        return wrapper

    def inject(
        self,
        task: t.Callable[
            t.Concatenate[d.Data, d.TaskParams],
            abc.Awaitable[d.TaskReturn],
        ],
    ) -> t.Callable[d.TaskParams, abc.Awaitable[d.TaskReturn]]:
        @functools.wraps(task)
        async def _wrapper(
            *args: d.TaskParams.args,
            **kwargs: d.TaskParams.kwargs,
        ):
            """Create and inject the dependency."""
            data = await self._constructor()
            return await task(data, *args, **kwargs)

        return _wrapper


def async_dependency(
    func: t.Callable[[], abc.Awaitable[d.Data]],
) -> AsyncInjector[d.Data]:
    return AsyncInjector(func)


if d.is_test_env():  # pragma: no cover
    import pytest

    GT_USER_ID = 1234
    FAKE_ID = 5678

    @pytest.mark.asyncio
    async def test_simple_usage() -> None:
        @async_dependency
        async def user_id():
            # perform an http request
            return GT_USER_ID

        @user_id.inject
        async def return_injected_value(user_id: int):
            return user_id

        assert await return_injected_value() == GT_USER_ID

    @pytest.mark.asyncio
    async def test_fake_value_context():
        @async_dependency
        async def user_id() -> int:
            return GT_USER_ID

        @user_id.inject
        async def get_user(user_id: int):
            return f"User-{user_id}"

        # Test normal behavior
        assert await get_user() == f"User-{GT_USER_ID}"

        # Test with fake value
        with user_id.fake_value(FAKE_ID) as fake_id:
            assert await get_user() == f"User-{FAKE_ID}"
            assert fake_id == FAKE_ID

        # Test restoration
        assert await get_user() == f"User-{GT_USER_ID}"

    @pytest.mark.asyncio
    async def test_faker_context():
        @async_dependency
        async def user_id() -> int:
            return GT_USER_ID

        @user_id.faker
        async def fake_user_id():
            return FAKE_ID

        @user_id.inject
        async def get_user(user_id: int):
            return f"User-{user_id}"

        # Test normal behavior
        assert await get_user() == f"User-{GT_USER_ID}"

        # Test with faker
        with fake_user_id():
            assert await get_user() == f"User-{FAKE_ID}"

        # Test restoration
        assert await get_user() == f"User-{GT_USER_ID}"

    @pytest.mark.asyncio
    async def test_nested_async_dependencies():
        @async_dependency
        async def get_token():
            return "token-123"

        @async_dependency
        @get_token.inject
        async def get_client(token: str):
            return f"client-{token}"

        @get_client.inject
        async def make_request(client: str):
            return f"request-with-{client}"

        result = await make_request()
        assert result == "request-with-client-token-123"

    @pytest.mark.asyncio
    async def test_exception_handling():
        msg = "Async error"

        @async_dependency
        async def failing_dependency():
            raise ValueError(msg)

        @failing_dependency.inject
        async def use_failing(dep: str):
            return dep

        with pytest.raises(ValueError, match=msg):
            await use_failing()

    @pytest.mark.asyncio
    async def test_multiple_fake_contexts():
        @async_dependency
        async def user_id() -> int:
            return GT_USER_ID

        @async_dependency
        async def api_key() -> str:
            return "prod-key"

        @user_id.inject
        @api_key.inject
        async def make_request(api_key: str, user_id: int):
            return f"request-{api_key}-{user_id}"

        # Test multiple fake contexts simultaneously
        with (
            user_id.fake_value(FAKE_ID) as fake_user,
            api_key.fake_value("test-key") as fake_key,
        ):
            result = await make_request()
            assert result == f"request-test-key-{FAKE_ID}"
            assert fake_user == FAKE_ID
            assert fake_key == "test-key"

    @pytest.mark.asyncio
    async def test_nested_fakers():
        @async_dependency
        async def config():
            return {"env": "prod"}

        @config.faker
        async def test_config():
            return {"env": "test"}

        @config.faker
        async def dev_config():
            return {"env": "dev"}

        @config.inject
        async def get_env(config: dict[str, str]):
            return config["env"]

        assert await get_env() == "prod"

        with test_config():
            assert await get_env() == "test"

            with dev_config():
                assert await get_env() == "dev"

                with test_config():
                    assert await get_env() == "test"

                assert await get_env() == "dev"

            assert await get_env() == "test"

        assert await get_env() == "prod"

    @pytest.mark.asyncio
    async def test_injection_with_args():
        @async_dependency
        async def prefix():
            return "test"

        @prefix.inject
        async def format_string(prefix: str, name: str, count: int = 1):
            return f"{prefix}-{name}-{count}"

        result = await format_string("user", count=5)
        assert result == "test-user-5"

    @pytest.mark.asyncio
    async def test_complex_async_workflow():
        events: list[str] = []

        @async_dependency
        async def tracker():
            events.append("create")
            return events

        @tracker.inject
        async def workflow(tracker: list[str]):
            tracker.append("start")
            return len(tracker)

        # Normal flow
        n = 2
        assert await workflow() == n
        assert events == ["create", "start"]

        # With fake
        events.clear()
        with tracker.fake_value([]) as fake_tracker:
            assert await workflow() == 1
            assert fake_tracker == ["start"]
            assert events == []
